@@ -73,12 +73,6 @@ macro_rules! construct_bigint {
                 one
             });
 
-            /// Minimum value
-            pub const MIN: $name = $name([0u64; $n_words]);
-
-            /// Maximum value
-            pub const MAX: $name = $name([::core::u64::MAX; $n_words]);
-
             /// Bit dimension
             pub const BITS: u32 = $n_words * 64;
 
@@ -113,13 +107,36 @@ macro_rules! construct_bigint {
             #[inline]
             pub fn bits_required(&self) -> usize {
                 let &$name(ref arr) = self;
-                for i in 1..$n_words {
-                    if arr[$n_words - i] > 0 {
-                        return (0x40 * ($n_words - i + 1))
-                            - arr[$n_words - i].leading_zeros() as usize;
+                if self.is_negative() {
+                    for i in 1..$n_words {
+                        if arr[$n_words - i] != ::core::u64::MAX {
+                            return (0x40 * ($n_words - i + 1)) + 1
+                                - (!arr[$n_words - i]).leading_zeros() as usize;
+                        }
                     }
+                    0x40 + 1 - (!arr[0]).leading_zeros() as usize
+                } else {
+                    for i in 1..$n_words {
+                        if arr[$n_words - i] > 0 {
+                            return (0x40 * ($n_words - i + 1))
+                                - arr[$n_words - i].leading_zeros() as usize;
+                        }
+                    }
+                    0x40 - arr[0].leading_zeros() as usize
                 }
-                0x40 - arr[0].leading_zeros() as usize
+            }
+
+            #[inline]
+            pub fn is_zero(&self) -> bool {
+                self[..] == [0; $n_words]
+            }
+
+            #[inline]
+            pub fn abs(self) -> $name {
+                if !self.is_negative() {
+                    return self;
+                }
+                (!self).wrapping_add($name::ONE)
             }
 
             /// Creates the integer value from a byte array using big-endian
@@ -228,6 +245,15 @@ macro_rules! construct_bigint {
                     return ($name(ret), sub_copy);
                 }
 
+                if sub_copy.is_negative() || shift_copy.is_negative() {
+                    assert!(
+                        sub_copy != $name::MIN || my_bits != $name::BITS as usize,
+                        "attempt to divide with overflow"
+                    );
+                }
+                sub_copy = sub_copy.abs();
+                shift_copy = shift_copy.abs();
+
                 // Bitwise long division
                 let mut shift = my_bits - your_bits;
                 shift_copy <<= shift;
@@ -243,7 +269,15 @@ macro_rules! construct_bigint {
                     shift -= 1;
                 }
 
-                ($name(ret), sub_copy)
+                let ret = match self.is_negative() == other.is_negative() {
+                    true => $name(ret),
+                    false => -$name(ret),
+                };
+                sub_copy = match self.is_negative() {
+                    true => -sub_copy,
+                    false => sub_copy,
+                };
+                (ret, sub_copy)
             }
             // same operation as in div_rem, not panicking when
             #[inline]
@@ -368,18 +402,39 @@ macro_rules! construct_bigint {
         impl Ord for $name {
             #[inline]
             fn cmp(&self, other: &$name) -> ::core::cmp::Ordering {
-                // We need to manually implement ordering because we use little-endian
-                // and the auto derive is a lexicographic ordering(i.e. memcmp)
-                // which with numbers is equivilant to big-endian
-                for i in 0..$n_words {
-                    if self[$n_words - 1 - i] < other[$n_words - 1 - i] {
-                        return ::core::cmp::Ordering::Less;
-                    }
-                    if self[$n_words - 1 - i] > other[$n_words - 1 - i] {
-                        return ::core::cmp::Ordering::Greater;
+                match (self.is_negative(), other.is_negative()) {
+                    (false, true) => ::core::cmp::Ordering::Greater,
+                    (true, false) => ::core::cmp::Ordering::Less,
+                    _ => {
+                        // We need to manually implement ordering because we use little-endian
+                        // and the auto derive is a lexicographic ordering(i.e. memcmp)
+                        // which with numbers is equivilant to big-endian
+                        for i in 0..$n_words {
+                            if self[$n_words - 1 - i] < other[$n_words - 1 - i] {
+                                return ::core::cmp::Ordering::Less;
+                            }
+                            if self[$n_words - 1 - i] > other[$n_words - 1 - i] {
+                                return ::core::cmp::Ordering::Greater;
+                            }
+                        }
+                        ::core::cmp::Ordering::Equal
                     }
                 }
-                ::core::cmp::Ordering::Equal
+            }
+        }
+
+        impl ::core::ops::Neg for $name {
+            type Output = Self;
+            fn neg(self) -> Self::Output {
+                assert!(
+                    $name::MIN != $name([::core::u64::MAX; $n_words]),
+                    "attempt to negate unsigned number"
+                );
+                assert!(
+                    self != $name::MIN,
+                    "attempt to negate the minimum value, which would overflow"
+                );
+                (!self).wrapping_add($name::ONE)
             }
         }
 
@@ -422,7 +477,8 @@ macro_rules! construct_bigint {
                 T: Into<$name>,
             {
                 let $name(ref me) = self;
-                let $name(ref you) = other.into();
+                let other = other.into();
+                let $name(ref you) = other;
                 let mut ret = [0u64; $n_words];
                 let mut carry = 0u64;
                 for i in 0..$n_words {
@@ -432,7 +488,12 @@ macro_rules! construct_bigint {
                     carry += flag as u64;
                     ret[i] = res;
                 }
-                (Self(ret), carry > 0)
+                let ret = Self(ret);
+                (
+                    ret,
+                    (self.is_negative() == other.is_negative())
+                        && (self.is_negative() != ret.is_negative()),
+                )
             }
 
             /// Wrapping (modular) addition. Computes `self + rhs`, wrapping around at
@@ -481,11 +542,7 @@ macro_rules! construct_bigint {
             where
                 T: Into<$name>,
             {
-                let other = other.into();
-                (
-                    self.wrapping_add(!other).wrapping_add($name::ONE),
-                    self < other,
-                )
+                self.overflowing_add((!other.into()).wrapping_add($name::ONE))
             }
 
             /// Wrapping (modular) subtraction. Computes `self - rhs`, wrapping around
@@ -525,46 +582,6 @@ macro_rules! construct_bigint {
                 }
             }
 
-            /// Calculates `self * rhs`
-            ///
-            /// Returns a tuple of the multiplication along with a boolean indicating
-            /// whether an arithmetic overflow would occur. If an overflow would
-            /// have occurred then the wrapped value is returned.
-            pub fn overflowing_mul<T>(self, other: T) -> ($name, bool)
-            where
-                T: Into<$name>,
-            {
-                let $name(ref me) = self;
-                let $name(ref you) = other.into();
-                let mut ret = [0u64; $n_words];
-                let mut overflow = false;
-                for i in 0..$n_words {
-                    let mut carry = 0u64;
-                    for j in 0..$n_words {
-                        if i + j >= $n_words {
-                            if me[i] > 0 && you[j] > 0 {
-                                overflow = true
-                            }
-                            continue;
-                        }
-                        let prev_carry = carry;
-                        let res = me[i] as u128 * you[j] as u128;
-                        carry = (res >> 64) as u64;
-                        let mul = (res & ::core::u64::MAX as u128) as u64;
-                        let (res, flag) = ret[i + j].overflowing_add(mul);
-                        carry += flag as u64;
-                        ret[i + j] = res;
-                        let (res, flag) = ret[i + j].overflowing_add(prev_carry);
-                        carry += flag as u64;
-                        ret[i + j] = res;
-                    }
-                    if carry > 0 {
-                        overflow = true
-                    }
-                }
-                (Self(ret), overflow)
-            }
-
             /// Wrapping (modular) multiplication. Computes `self * rhs`, wrapping
             /// around at the boundary of the type.
             pub fn wrapping_mul<T>(self, other: T) -> $name
@@ -573,6 +590,36 @@ macro_rules! construct_bigint {
             {
                 self.overflowing_mul(other).0
             }
+
+            /// Checked shift left. Computes self << rhs,
+            /// returning None if rhs is larger than or equal to the number of bits in self.
+            pub fn checked_shl(self, rhs: u32) -> Option<$name> {
+                match rhs < Self::BITS {
+                    true => Some(self << (rhs as usize)),
+                    false => None,
+                }
+            }
+
+            /// Checked shift right. Computes self >> rhs,
+            /// returning None if rhs is larger than or equal to the number of bits in self.
+            pub fn checked_shr(self, rhs: u32) -> Option<$name> {
+                match rhs < Self::BITS {
+                    true => Some(self >> (rhs as usize)),
+                    false => None,
+                }
+            }
+
+            /// Wrapping (modular) negation. Computes -self,
+            /// wrapping around at the boundary of the type.
+            /// Since unsigned types do not have negative equivalents
+            /// all applications of this function will wrap (except for -0).
+            /// For values smaller than the corresponding signed type's maximum
+            /// the result is the same as casting the corresponding signed value.
+            /// Any larger values are equivalent to MAX + 1 - (val - MAX - 1)
+            /// where MAX is the corresponding signed type's maximum.
+            pub fn wrapping_neg(self) -> $name {
+                !self + Self::ONE
+            }
         }
 
         impl<T> ::core::ops::Add<T> for $name
@@ -807,9 +854,13 @@ macro_rules! construct_bigint {
                         ret[i - word_shift] += original[i + 1] << (64 - bit_shift);
                     }
                 }
+                if self.is_negative() {
+                    ret[$n_words - 1] |= 0x8000_0000_0000_0000
+                }
                 $name(ret)
             }
         }
+
         impl ::core::ops::ShrAssign<usize> for $name {
             #[inline]
             fn shr_assign(&mut self, rhs: usize) {
@@ -1060,9 +1111,222 @@ macro_rules! construct_bigint {
     };
 }
 
+macro_rules! construct_signed_bigint_methods {
+    ( $ name: ident, $ n_words: expr ) => {
+        impl From<i8> for $name {
+            fn from(init: i8) -> $name {
+                let bytes = init.to_le_bytes();
+                let mut ret = [if init.is_negative() {
+                    ::core::u8::MAX
+                } else {
+                    0
+                }; $n_words * 8];
+                for i in 0..bytes.len() {
+                    ret[i] = bytes[i]
+                }
+                $name::from_le_bytes(ret)
+            }
+        }
+
+        impl From<i16> for $name {
+            fn from(init: i16) -> $name {
+                let bytes = init.to_le_bytes();
+                let mut ret = [if init.is_negative() {
+                    ::core::u8::MAX
+                } else {
+                    0
+                }; $n_words * 8];
+                for i in 0..bytes.len() {
+                    ret[i] = bytes[i]
+                }
+                $name::from_le_bytes(ret)
+            }
+        }
+
+        impl From<i32> for $name {
+            fn from(init: i32) -> $name {
+                let bytes = init.to_le_bytes();
+                let mut ret = [if init.is_negative() {
+                    ::core::u8::MAX
+                } else {
+                    0
+                }; $n_words * 8];
+                for i in 0..bytes.len() {
+                    ret[i] = bytes[i]
+                }
+                $name::from_le_bytes(ret)
+            }
+        }
+
+        impl From<i64> for $name {
+            fn from(init: i64) -> $name {
+                let bytes = init.to_le_bytes();
+                let mut ret = [if init.is_negative() {
+                    ::core::u8::MAX
+                } else {
+                    0
+                }; $n_words * 8];
+                for i in 0..bytes.len() {
+                    ret[i] = bytes[i]
+                }
+                $name::from_le_bytes(ret)
+            }
+        }
+
+        impl From<i128> for $name {
+            fn from(init: i128) -> $name {
+                let bytes = init.to_le_bytes();
+                let mut ret = [if init.is_negative() {
+                    ::core::u8::MAX
+                } else {
+                    0
+                }; $n_words * 8];
+                for i in 0..bytes.len() {
+                    ret[i] = bytes[i]
+                }
+                $name::from_le_bytes(ret)
+            }
+        }
+
+        impl $name {
+            /// Minimum value
+            pub const MIN: $name = {
+                let mut min = [0u64; $n_words];
+                min[$n_words - 1] = 0x8000_0000_0000_0000;
+                $name(min)
+            };
+
+            /// Maximum value
+            pub const MAX: $name = {
+                let mut max = [::core::u64::MAX; $n_words];
+                max[$n_words - 1] = ::core::u64::MAX >> 1;
+                $name(max)
+            };
+
+            #[inline]
+            pub fn is_positive(&self) -> bool {
+                !self.is_zero()
+                    && self[($name::INNER_LEN - 1) as usize] & 0x8000_0000_0000_0000 == 0
+            }
+
+            #[inline]
+            pub fn is_negative(&self) -> bool {
+                !self.is_zero() && !self.is_positive()
+            }
+
+            /// Calculates `self * rhs`
+            ///
+            /// Returns a tuple of the multiplication along with a boolean indicating
+            /// whether an arithmetic overflow would occur. If an overflow would
+            /// have occurred then the wrapped value is returned.
+            pub fn overflowing_mul<T>(self, other: T) -> ($name, bool)
+            where
+                T: Into<$name>,
+            {
+                let sub = if self != $name::MIN { -self } else { self };
+                let mut p_high = $name::ZERO;
+                let mut p_low = other.into();
+                let mut prev = false;
+                for _i in 0..$name::BITS {
+                    let p_low_trailing_bit = (p_low[0] & 1) != 0;
+                    p_high = match (p_low_trailing_bit, prev) {
+                        (false, true) => p_high.wrapping_add(self),
+                        (true, false) => p_high.wrapping_add(sub),
+                        _ => p_high,
+                    };
+                    prev = p_low_trailing_bit;
+                    p_low >>= 1;
+                    p_low = match p_high[0] & 1 {
+                        0 => p_low & $name::MAX,
+                        _ => p_low | $name::MIN,
+                    };
+                    p_high >>= 1;
+                }
+                let negative_overflow =
+                    p_low.is_negative() && p_high != $name([::core::u64::MAX; $n_words]);
+                let positive_overflow = !p_low.is_negative() && p_high != $name::ZERO;
+                (p_low, negative_overflow || positive_overflow)
+            }
+        }
+    };
+}
+
+macro_rules! construct_unsigned_bigint_methods {
+    ( $ name: ident, $ n_words: expr ) => {
+        impl $name {
+            /// Minimum value
+            pub const MIN: $name = $name([0u64; $n_words]);
+
+            /// Maximum value
+            pub const MAX: $name = $name([::core::u64::MAX; $n_words]);
+
+            #[inline]
+            pub fn is_positive(&self) -> bool {
+                !self.is_zero()
+            }
+
+            #[inline]
+            pub fn is_negative(&self) -> bool {
+                false
+            }
+
+            /// Calculates `self * rhs`
+            ///
+            /// Returns a tuple of the multiplication along with a boolean indicating
+            /// whether an arithmetic overflow would occur. If an overflow would
+            /// have occurred then the wrapped value is returned.
+            pub fn overflowing_mul<T>(self, other: T) -> ($name, bool)
+            where
+                T: Into<$name>,
+            {
+                let $name(ref me) = self;
+                let $name(ref you) = other.into();
+                let mut ret = [0u64; $n_words];
+                let mut overflow = false;
+                for i in 0..$n_words {
+                    let mut carry = 0u64;
+                    for j in 0..$n_words {
+                        if i + j >= $n_words {
+                            if me[i] > 0 && you[j] > 0 {
+                                overflow = true
+                            }
+                            continue;
+                        }
+                        let prev_carry = carry;
+                        let res = me[i] as u128 * you[j] as u128;
+                        carry = (res >> 64) as u64;
+                        let mul = (res & ::core::u64::MAX as u128) as u64;
+                        let (res, flag) = ret[i + j].overflowing_add(mul);
+                        carry += flag as u64;
+                        ret[i + j] = res;
+                        let (res, flag) = ret[i + j].overflowing_add(prev_carry);
+                        carry += flag as u64;
+                        ret[i + j] = res;
+                    }
+                    if carry > 0 {
+                        overflow = true
+                    }
+                }
+                (Self(ret), overflow)
+            }
+        }
+    };
+}
+
+construct_bigint!(i256, 4);
+construct_bigint!(i512, 8);
+construct_bigint!(i1024, 16);
 construct_bigint!(u256, 4);
 construct_bigint!(u512, 8);
 construct_bigint!(u1024, 16);
+
+construct_unsigned_bigint_methods!(u256, 4);
+construct_unsigned_bigint_methods!(u512, 8);
+construct_unsigned_bigint_methods!(u1024, 16);
+construct_signed_bigint_methods!(i256, 4);
+construct_signed_bigint_methods!(i512, 8);
+construct_signed_bigint_methods!(i1024, 16);
+
 #[cfg(test)]
 mod tests {
     #![allow(unused)]
@@ -1070,6 +1334,7 @@ mod tests {
     use super::*;
 
     construct_bigint!(Uint128, 2);
+    construct_unsigned_bigint_methods!(Uint128, 2);
 
     #[test]
     fn u256_bits_test() {
@@ -1597,4 +1862,163 @@ mod tests {
         )
         .is_err()); // invalid length
     }
+
+    #[test]
+    fn i256_test() {
+        let x = i256::from(1);
+        let y = i256::from(1);
+        assert_eq!(x.checked_add(y), Some(i256::from(2)));
+    }
+
+    #[test]
+    fn i256_is_positive_test() {
+        assert_eq!(true, i256::from(1).is_positive());
+        assert_eq!(false, i256::from(-1).is_positive());
+        assert_eq!(false, i256::from(0).is_positive());
+        assert_eq!(true, i256::MAX.is_positive());
+        assert_eq!(false, i256::MIN.is_positive());
+        assert_eq!(true, i256::MIN.is_negative());
+    }
+
+    #[test]
+    fn i256_add_test() {
+        assert_eq!(
+            (i256::from(3), false),
+            i256::from(1).overflowing_add(i256::from(2))
+        );
+        assert_eq!(
+            (i256::from(1), false),
+            i256::from(-1).overflowing_add(i256::from(2))
+        );
+        assert_eq!(
+            (i256::from(-2), false),
+            i256::from(-1).overflowing_add(i256::from(-1))
+        );
+        assert_eq!(
+            (i256::from(0), false),
+            i256::from(0).overflowing_add(i256::from(0))
+        );
+        assert_eq!((i256::MIN, true), i256::from(1).overflowing_add(i256::MAX));
+    }
+
+    #[test]
+    fn i256_sub_test() {
+        assert_eq!(
+            (i256::from(-1), false),
+            i256::from(1).overflowing_sub(i256::from(2))
+        );
+        assert_eq!(
+            (i256::from(1), false),
+            i256::from(3).overflowing_sub(i256::from(2))
+        );
+        assert_eq!(
+            (i256::from(-3), false),
+            i256::from(-4).overflowing_sub(i256::from(-1))
+        );
+        assert_eq!(
+            (i256::from(0), false),
+            i256::from(0).overflowing_add(i256::from(0))
+        );
+        assert_eq!((i256::MIN, false), i256::from(0).overflowing_sub(i256::MIN));
+        assert_eq!((-i256::ONE, false), i256::MAX.overflowing_sub(i256::MIN));
+        assert_eq!(
+            (i256::MAX, true),
+            (-i256::from(2)).overflowing_sub(i256::MAX)
+        );
+    }
+
+    #[test]
+    fn i256_neg_test() {
+        assert_eq!(i256::from(1), -i256::from(-1));
+        assert_eq!(i256::from(-1), -i256::from(1));
+        assert_eq!(i256::from(0), -i256::from(0));
+        assert_eq!(i256::MIN + 1, -i256::MAX);
+    }
+
+    #[test]
+    #[should_panic]
+    fn i256_neg_min_test() {
+        assert_eq!(-i256::MIN, -i256::MIN);
+    }
+
+    #[test]
+    fn i256_mul_test() {
+        assert_eq!(
+            (i256::from(-12), false),
+            i256::from(3).overflowing_mul(i256::from(-4))
+        );
+        assert_eq!(
+            (i256::from(6), false),
+            i256::from(2).overflowing_mul(i256::from(3))
+        );
+        assert_eq!(
+            (i256::from(30), false),
+            i256::from(-6).overflowing_mul(i256::from(-5))
+        );
+        assert_eq!(
+            (i256::from(-2), true),
+            i256::MAX.overflowing_mul(i256::from(2))
+        );
+        assert_eq!((i256::ZERO, true), i256::MIN.overflowing_mul(i256::from(2)));
+        assert_eq!((i256::ONE, true), i256::MAX.overflowing_mul(i256::MAX));
+    }
+
+    #[test]
+    fn i256_arithmetic_shr_test() {
+        assert_eq!(i256::from(-1), i256::from(-1) >> 1);
+        assert_eq!(i256::from(-1), i256::from(-2) >> 1);
+        assert_eq!(i256::from(1), i256::from(2) >> 1);
+        assert_eq!(i256::from(1), i256::from(2) >> 1);
+        assert_eq!(i256::from(0), i256::from(1) >> 1);
+    }
+
+    #[test]
+    fn i256_bits_required_test() {
+        assert_eq!(i256::from(255u64).bits_required(), 8);
+        assert_eq!(i256::from(256u64).bits_required(), 9);
+        assert_eq!(i256::from(300u64).bits_required(), 9);
+        assert_eq!(i256::from(60000u64).bits_required(), 16);
+        assert_eq!(i256::from(70000u64).bits_required(), 17);
+        assert_eq!(i256::from(-128i64).bits_required(), 8);
+        assert_eq!(i256::from(-129i128).bits_required(), 9);
+        assert_eq!(i256::from(0i32).bits_required(), 0);
+        assert_eq!(i256::from(-1i16).bits_required(), 1);
+        assert_eq!(i256::from(-2i64).bits_required(), 2);
+        assert_eq!(i256::MIN.bits_required(), 256);
+        assert_eq!(i256::MAX.bits_required(), 255);
+    }
+
+    #[test]
+    fn i256_div_test() {
+        assert_eq!(
+            (i256::from(3), i256::from(1)),
+            i256::from(7).div_rem_checked(i256::from(2i32)).unwrap()
+        );
+        assert_eq!(
+            (i256::from(-3), i256::from(1)),
+            i256::from(7).div_rem_checked(i256::from(-2i128)).unwrap()
+        );
+        assert_eq!(
+            (i256::from(-3), i256::from(-1)),
+            i256::from(-7).div_rem_checked(i256::from(2)).unwrap()
+        );
+        assert_eq!(
+            (i256::from(3), i256::from(-1)),
+            i256::from(-7).div_rem_checked(i256::from(-2)).unwrap()
+        );
+        let res = std::panic::catch_unwind(|| i256::div_rem(i256::MAX, i256::ZERO));
+        assert!(res.is_err());
+        let res = std::panic::catch_unwind(|| i256::div_rem(i256::MIN, i256::from(-1)));
+        assert!(res.is_err());
+    }
+
+    #[test]
+    fn i256_cmp_test() {
+        assert!(i256::ZERO < i256::ONE);
+        assert!(-i256::ONE < i256::ZERO);
+        assert!(i256::MIN < i256::MAX);
+        assert!(i256::MIN < i256::ZERO);
+        assert!(i256::from(200) < i256::from(10000000));
+        assert!(i256::from(-3) < i256::from(87));
+    }
 }
